local addonName, AT = ...
local DEBUG = false

local Constants = AT.Constants
local Utils = AT.Utils

local PROFILE_TYPE = AT.Enums.PROFILE_TYPE

local defaultGlobalSettings = {
  showWelcomeMsg = true,
  printLoadedProfileAtStartup = true,

  testFramesEnabled = false,
  testFramesShowMouseover = true,
}

local defaultGeneralSettings = {
  buffStolenChatMessage = 1,
  frameDragWithModifierKey = false,
  frameEnableMouse = true,

  soundTargetEnabled = true,
  soundTargetFile = 'Interface\\AddOns\\SpellstealMaster\\sounds\\371753__problematist__dhumm2.ogg',
  soundTargetChannel = 'SFX',
  soundTargetMinTimeGap = 1,
  soundFocusEnabled = true,
  soundFocusFile = 'Interface\\AddOns\\SpellstealMaster\\sounds\\371753__problematist__dhumm2.ogg',
  soundFocusChannel = 'SFX',
  soundFocusMinTimeGap = 1,
  soundMouseoverEnabled = true,
  soundMouseoverFile = 'Interface\\AddOns\\SpellstealMaster\\sounds\\262958__unicornaphobist__glass-tap.wav',
  soundMouseoverChannel = 'SFX',
  soundMouseoverMinTimeGap = 0.1,
  soundOverviewEnabled = true,
  soundOverviewFile = 'Interface\\AddOns\\SpellstealMaster\\sounds\\371752__problematist__dhumm.ogg',
  soundOverviewChannel = 'SFX',
  soundOverviewMinTimeGap = 2.5,
  soundNpEnabled = true,
  soundNpFile = 'Interface\\AddOns\\SpellstealMaster\\sounds\\371752__problematist__dhumm.ogg',
  soundNpChannel = 'SFX',
  soundNpMinTimeGap = 2.5,

  colorBg = { r = 0, g = 0, b = 0, a = 0.55 },
  colorBorder = { r = 0.18, g = 0.65, b = 1, a = 1 },
  colorHeadline = { r = 0.18, g = 0.65, b = 1, a = 1 },
  colorUnitName = { r = 0.98, g = 0.72, b = 0.72, a = 1 },
  colorUnitOrigin = { r = 0.52, g = 0.52, b = 0.52, a = 1 },
  colorBuffName = { r = 0.58, g = 1, b = 0.68, a = 1 },
  colorBuffCount = { r = 0.69, g = 0.69, b = 0.69, a = 1 },
  colorBuffExpTime = { r = 0.69, g = 0.69, b = 0.69, a = 1 },
  alphaNoMana = 0.33,
  alphaOutOfRange = 0.66,

  frameTargetEnabled = true,
  frameTargetShowBorder = true,
  frameTargetStrata = 'MEDIUM',
  frameTargetScale = 1,
  frameTargetPivot = 'CENTER',
  frameTargetAnchor = 'CENTER',
  frameTargetOffsetX = 0,
  frameTargetOffsetY = 100,

  frameFocusEnabled = true,
  frameFocusShowBorder = true,
  frameFocusStrata = 'MEDIUM',
  frameFocusScale = 1,
  frameFocusPivot = 'CENTER',
  frameFocusAnchor = 'CENTER',
  frameFocusOffsetX = 200,
  frameFocusOffsetY = 0,

  frameOverviewEnabled = true,
  frameOverviewShowBorder = true,
  frameOverviewStrata = 'MEDIUM',
  frameOverviewScale = 1,
  frameOverviewMinWidth = 0,
  frameOverviewMaxWidth = 150,
  frameOverviewMinHeight = 0,
  frameOverviewMaxHeight = 200,
  frameOverviewPivot = 'CENTER',
  frameOverviewAnchor = 'CENTER',
  frameOverviewOffsetX = -200,
  frameOverviewOffsetY = 0,
  frameOverviewDisplayUnitOrigin = true,

  frameMouseoverEnabled = true,
  frameMouseoverShowBorder = true,
  frameMouseoverStrata = 'TOOLTIP',
  frameMouseoverScale = 1,
  frameMouseoverOffsetX = 20,
  frameMouseoverOffsetY = 0,

  npEnabled = true,
  npStrata = 'MEDIUM',
  npOffsetX = 0,
  npOffsetY = 39,
  npAutoMaxWidth = true,
  npMaxWidth = 170,
  npIconWidth = 20,
  npIconHeight = 20,
  npIconBorderSize = 2,
  npIconBorderColor = { r = 1, g = 1, b = 0, a = 1 },
  npIconSpacing = 4,
  npIconOrientation = 1,
  npIconKeepOpaque = false
}

local defaultFilterSettings = {
  filterIgnoreGroupBuffs = true,
  filterGlobalCsv = '',
  filterGlobalCsvIsBlacklist = true,
  filterContextualCsv = '',
  filterContextualCsvIsBlacklist = true,
  filterContextualLua = '',

  filterImportantMinExpirationTime = 2,

  filterOverviewMinExpirationTime = 5,
  filterOverviewShowTarget = false,
  filterOverviewShowFocus = false,
  filterOverviewShowNameplates = true,
  filterOverviewShowBosses = true,
  filterOverviewShowPetTarget = true,
  filterOverviewShowGroupMemberTargets = true,
  filterOverviewShowGroupMemberPetTargets = true,
}

local Settings = {}
Settings.Global = {}
Settings.General = {}
Settings.Filter = {}
Settings.OnProfileChanged = {
  listeners = {},
  Add = function(self, aTable, aFunc, aProfileTypeFlags) 
    tinsert(self.listeners, {
      table = aTable,
      func = aFunc,
      flags = aProfileTypeFlags
    })
  end,
  Broadcast = function(self, aProfileType)
    for _, v in ipairs(self.listeners) do
      if Enums:HasFlag(v.flags, aProfileType) then
        v.func(v.table, aProfileType)
      end
    end
  end
}
Settings.isInitialized = false

AT.Settings = Settings

function Settings:Init()
  setmetatable(Settings, { __index = Settings.Get, __newindex = Settings.Set })

  if SPELLSTEALMASTER_SETTINGS_GLOBAL == nil then
    SPELLSTEALMASTER_SETTINGS_GLOBAL = {}
  end
  setmetatable(SPELLSTEALMASTER_SETTINGS_GLOBAL, {__index = defaultGlobalSettings})
  self.Global = SPELLSTEALMASTER_SETTINGS_GLOBAL

  self:LoadCurrentProfile(PROFILE_TYPE.GENERAL)
  self:LoadCurrentProfile(PROFILE_TYPE.FILTER)

  self.isInitialized = true
end

function Settings:Get(aKey)
  local val = Settings.General[aKey]
  if val == nil then
    val = Settings.Filter[aKey]
  end
  if val == nil then
    val = Settings.Global[aKey]
  end
  if DEBUG and val == nil then
    error('Setting '..aKey..' does not exist')
  end
  return val
end

function Settings:Set(aKey, aValue)
  if Settings.General[aKey] ~= nil then
    Settings.General[aKey] = aValue
  elseif Settings.Filter[aKey] ~= nil then
    Settings.Filter[aKey] = aValue
  elseif Settings.Global[aKey] ~= nil then
    Settings.Global[aKey] = aValue
  elseif DEBUG then
    error('Setting '..aKey..' does not exist')
  end
end

function Settings:CreateProfile(aProfileType, aProfileName)
  local gt = self:GetGlobalTable(aProfileType)

  if self:ProfileExists(aProfileType, aProfileName) then
    Utils:Print(string.format("profile %s (type: %s) already exists", self:DyeProfileName(aProfileName), self:ProfileTypeToStrColored(aProfileType)))
    return
  end

  gt.profiles[aProfileName] = {}
  Utils:Print(string.format("created profile %s (type: %s)", self:DyeProfileName(aProfileName), self:ProfileTypeToStrColored(aProfileType)))
  self:LoadProfile(aProfileType, aProfileName)
end

function Settings:CopyProfile(aProfileType, aFromProfileName, aToProfileName, aLoadCreated)
  local gt = self:GetGlobalTable(aProfileType)

  if not self:ProfileExists(aProfileType, aFromProfileName, true) then
    return false
  end

  if self:ProfileExists(aProfileType, aToProfileName) then
    Utils:Print(string.format("profile %s (type: %s) already exists", self:DyeProfileName(aToProfileName), self:ProfileTypeToStrColored(aProfileType)))
    return false
  end

  gt.profiles[aToProfileName] = Utils:TableDeepCopy(gt.profiles[aFromProfileName])
  Utils:Print(string.format("created profile %s (type: %s) from profile %s",
    self:DyeProfileName(aToProfileName), self:ProfileTypeToStrColored(aProfileType), self:DyeProfileName(aFromProfileName)))
  if aLoadCreated == nil or aLoadCreated == true then
    self:LoadProfile(aProfileType, aToProfileName)
  end

  return true
end

function Settings:RenameProfile(aProfileType, aFromProfileName, aToProfileName)
  if self:CopyProfile(aProfileType, aFromProfileName, aToProfileName, false) then
    self:DeleteProfile(aProfileType, aFromProfileName)
  end
end

function Settings:LoadProfile(aProfileType, aProfileName)
  local gt = self:GetGlobalTable(aProfileType)

  if not self:ProfileExists(aProfileType, aProfileName, true) then
    return
  end

  gt.currentProfile = aProfileName

  local pt = gt.profiles[aProfileName]
  setmetatable(pt, {__index = self:GetDefaultSettings(aProfileType)})

  if aProfileType == PROFILE_TYPE.GENERAL then
    self.General = pt
  else
    self.Filter = pt
  end

  if self.isInitialized or (Utils:TableLen(gt.profiles) > 1 and Settings.printLoadedProfileAtStartup) then
    Utils:Print(string.format("loaded profile %s (type: %s)", self:DyeProfileName(aProfileName), self:ProfileTypeToStrColored(aProfileType)))
  end

  self.OnProfileChanged:Broadcast(aProfileType)
end

function Settings:LoadCurrentProfile(aProfileType)
  local gt = self:GetGlobalTable(aProfileType)

  if not self:ProfileExists(aProfileType, gt.currentProfile) then
    gt.currentProfile = 'default'

    if not self:ProfileExists(aProfileType, gt.currentProfile) then
      self:CreateProfile(aProfileType, gt.currentProfile, false)
      return
    end
  end

  self:LoadProfile(aProfileType, gt.currentProfile)
end

function Settings:LoadNextProfile(aProfileType)
  local gt = self:GetGlobalTable(aProfileType)

  local profiles = {}
  for k, _ in pairs(gt.profiles) do
    tinsert(profiles, k)
  end
  table.sort(profiles, function(a, b) return a < b end)

  local currentIndex = Utils:TableValueIndex(profiles, gt.currentProfile)
  local nextProfile = profiles[currentIndex < #profiles and currentIndex + 1 or 1]

  self:LoadProfile(aProfileType, nextProfile)
end

function Settings:DeleteProfile(aProfileType, aProfileName)
  if not self:ProfileExists(aProfileType, aProfileName, true) then
    return
  end

  local gt = self:GetGlobalTable(aProfileType)
  gt.profiles[aProfileName] = nil

  Utils:Print(string.format("deleted profile %s (type: %s)", self:DyeProfileName(aProfileName), self:ProfileTypeToStrColored(aProfileType)))

  if aProfileName == gt.currentProfile then
    self:LoadCurrentProfile(aProfileType)
  end
end

function Settings:ProfileExists(aProfileType, aProfileName, aPrintOnError)
  local gt = self:GetGlobalTable(aProfileType)

  if not Utils:StrIsNilOrEmpty(gt.profiles[aProfileName]) then
    return true
  end

  if aPrintOnError then
    Utils:Print(string.format("profile %s (type: %s) does not exists", self:DyeProfileName(aProfileName), self:ProfileTypeToStrColored(aProfileType)))
  end

  return false
end

function Settings:ListProfiles(aProfileType)
  local strBuilder = {}

  if aProfileType == nil or aProfileType == PROFILE_TYPE.GENERAL then
    tinsert(strBuilder, '\n')
    tinsert(strBuilder, "General profiles:")
    for k, _ in pairs(self:GetGlobalTable(PROFILE_TYPE.GENERAL).profiles) do
      tinsert(strBuilder, '\n- '..k)
    end
  end

  if aProfileType == nil or aProfileType == PROFILE_TYPE.FILTER then
    tinsert(strBuilder, '\n')
    tinsert(strBuilder, "Filter profiles:")
    for k, _ in pairs(self:GetGlobalTable(PROFILE_TYPE.FILTER).profiles) do
      tinsert(strBuilder, '\n- '..k)
    end
  end

  Utils:Print(table.concat(strBuilder))
end

function Settings:GetGlobalTable(aProfileType)
  local t
  if aProfileType == PROFILE_TYPE.GENERAL then
    if SPELLSTEALMASTER_SETTINGS_GENERAL == nil then
      SPELLSTEALMASTER_SETTINGS_GENERAL = {}
    end
    t = SPELLSTEALMASTER_SETTINGS_GENERAL
  else
    if SPELLSTEALMASTER_SETTINGS_FILTER == nil then
      SPELLSTEALMASTER_SETTINGS_FILTER = {}
    end
    t = SPELLSTEALMASTER_SETTINGS_FILTER
  end

  if Utils:StrIsNilOrEmpty(t.currentProfile) then
    t.currentProfile = 'default'
  end
  if t.profiles == nil then
    t.profiles = {}
  end

  return t
end

function Settings:GetDefaultSettings(aProfileType)
  return aProfileType == PROFILE_TYPE.GENERAL
    and defaultGeneralSettings
    or defaultFilterSettings
end

function Settings:ProfileTypeToStr(aProfileType)
  return aProfileType == PROFILE_TYPE.GENERAL
    and "general"
    or "filter"
end

function Settings:ProfileTypeToStrColored(aProfileType)
  return string.format("|c%s%s|r", Constants.COLOR_PROFILE_TYPE_HEX, self:ProfileTypeToStr(aProfileType))
end

function Settings:DyeProfileName(aProfileName)
  return string.format("|c%s%s|r", Constants.COLOR_PROFILE_NAME_HEX, aProfileName)
end
